"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 *  Copyright (c) 2020 DeltaXML Ltd. and others.
 *
 *  Contributors:
 *  DeltaXML Ltd. - xsltTokenDiagnostics
 */
const vscode = require("vscode");
const xslLexer_1 = require("./xslLexer");
const xpLexer_1 = require("./xpLexer");
const functionData_1 = require("./functionData");
const xsltTokenDiagnostics_1 = require("./xsltTokenDiagnostics");
const url = require("url");
var TagType;
(function (TagType) {
    TagType[TagType["XSLTstart"] = 0] = "XSLTstart";
    TagType[TagType["XMLstart"] = 1] = "XMLstart";
    TagType[TagType["XSLTvar"] = 2] = "XSLTvar";
    TagType[TagType["Start"] = 3] = "Start";
    TagType[TagType["NonStart"] = 4] = "NonStart";
})(TagType || (TagType = {}));
var AttributeType;
(function (AttributeType) {
    AttributeType[AttributeType["None"] = 0] = "None";
    AttributeType[AttributeType["Variable"] = 1] = "Variable";
    AttributeType[AttributeType["InstructionName"] = 2] = "InstructionName";
    AttributeType[AttributeType["InstructionMode"] = 3] = "InstructionMode";
    AttributeType[AttributeType["UseAttributeSets"] = 4] = "UseAttributeSets";
    AttributeType[AttributeType["ExcludeResultPrefixes"] = 5] = "ExcludeResultPrefixes";
})(AttributeType || (AttributeType = {}));
class XsltTokenDefinitions {
    static createLocationFromInstrcution(instruction, document) {
        if (instruction) {
            let s = url.pathToFileURL;
            let uri = (instruction === null || instruction === void 0 ? void 0 : instruction.href) ? vscode.Uri.parse(url.pathToFileURL(instruction.href).toString()) : document.uri;
            let startPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter);
            let endPos = new vscode.Position(instruction.token.line, instruction.token.startCharacter + instruction.token.length);
            return new vscode.Location(uri, new vscode.Range(startPos, endPos));
        }
    }
    static resolveFunctionName(xmlnsPrefixes, xmlnsData, token) {
        let parts = token.value.split('#');
        let arity = Number.parseInt(parts[1]);
        let name = parts[0];
        return { name, arity };
    }
    static resolveXPathVariableReference(document, importedVariables, token, xpathVariableCurrentlyBeingDefined, inScopeXPathVariablesList, xpathStack, inScopeVariablesList, elementStack) {
        let fullVarName = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(token.line, token, document);
        let varName = fullVarName.substr(1);
        let resolved = null;
        let globalVariable = null;
        resolved = this.resolveVariableName(inScopeXPathVariablesList, varName, xpathVariableCurrentlyBeingDefined, globalVariable);
        if (!resolved) {
            resolved = this.resolveStackVariableName(xpathStack, varName);
        }
        if (!resolved) {
            resolved = this.resolveVariableName(inScopeVariablesList, varName, false, globalVariable);
        }
        if (!resolved) {
            resolved = this.resolveStackVariableName(elementStack, varName);
        }
        if (!resolved) {
            let instruction = importedVariables.get(varName);
            if (instruction) {
                resolved = { name: instruction.name, token: instruction.token, uri: instruction.href };
            }
        }
        return resolved;
    }
    static resolveVariableName(variableList, varName, xpathVariableCurrentlyBeingDefined, globalXsltVariable) {
        let resolved = null;
        let decrementedLength = variableList.length - 1;
        let globalVariableName = globalXsltVariable === null || globalXsltVariable === void 0 ? void 0 : globalXsltVariable.name;
        // last items in list of declared parameters must be resolved first:
        for (let i = decrementedLength; i > -1; i--) {
            let data = variableList[i];
            if (xpathVariableCurrentlyBeingDefined && i === decrementedLength) {
                // do nothing: we skip last item in list as it's currently being defined
            }
            else if (data.name === varName && globalVariableName !== data.name) {
                resolved = data;
                break;
            }
        }
        return resolved;
    }
    static resolveStackVariableName(elementStack, varName) {
        let resolved = null;
        let globalXsltVariable = null;
        for (let i = elementStack.length - 1; i > -1; i--) {
            let inheritedVariables = elementStack[i].variables;
            let xpathBeingDefinedInit = elementStack[i].xpathVariableCurrentlyBeingDefined;
            let xpathBeingDefined = !(xpathBeingDefinedInit === undefined || xpathBeingDefinedInit === false);
            if (i === 1) {
                // at the level of a global variable declaration
                let elementData = elementStack[i];
                let currentVar = elementData.currentVariable;
                if (currentVar) {
                    // must be inside a global variable declaration - keep this:
                    globalXsltVariable = currentVar;
                }
            }
            resolved = this.resolveVariableName(inheritedVariables, varName, xpathBeingDefined, globalXsltVariable);
            if (resolved) {
                break;
            }
        }
        return resolved;
    }
    static findMatchingDefintion(globalInstructionData, importedInstructionData, name, type, arity) {
        let findFunction = type === xslLexer_1.GlobalInstructionType.Function;
        let found = globalInstructionData.find((instruction) => {
            if (type !== instruction.type) {
                return;
            }
            else if (findFunction && arity) {
                return instruction.name === name && instruction.idNumber === arity;
            }
            else {
                return instruction.name === name;
            }
        });
        if (!found) {
            found = importedInstructionData.find((instruction) => {
                if (type !== instruction.type) {
                    return;
                }
                else if (findFunction && arity) {
                    return instruction.name === name && instruction.idNumber === arity;
                }
                else {
                    return instruction.name === name;
                }
            });
        }
        return found;
    }
}
exports.XsltTokenDefinitions = XsltTokenDefinitions;
XsltTokenDefinitions.xsltStartTokenNumber = xslLexer_1.XslLexer.getXsltStartTokenNumber();
XsltTokenDefinitions.xslVariable = ['xsl:variable', 'xsl:param'];
XsltTokenDefinitions.xslInclude = 'xsl:include';
XsltTokenDefinitions.xslImport = 'xsl:import';
XsltTokenDefinitions.xmlChars = ['lt', 'gt', 'quot', 'apos', 'amp'];
XsltTokenDefinitions.xslFunction = 'xsl:function';
XsltTokenDefinitions.xslNameAtt = 'name';
XsltTokenDefinitions.xslModeAtt = 'mode';
XsltTokenDefinitions.useAttSet = 'use-attribute-sets';
XsltTokenDefinitions.xslUseAttSet = 'xsl:use-attribute-sets';
XsltTokenDefinitions.excludePrefixes = 'exclude-result-prefixes';
XsltTokenDefinitions.xslExcludePrefixes = 'xsl:exclude-result-prefixes';
XsltTokenDefinitions.findDefinition = (isXSLT, document, allTokens, globalInstructionData, importedInstructionData, position) => {
    var _a, _b;
    let lineNumber = -1;
    let resultLocation;
    let inScopeVariablesList = [];
    let xpathVariableCurrentlyBeingDefined = false;
    let elementStack = [];
    let inScopeXPathVariablesList = [];
    let anonymousFunctionParamList = [];
    let xpathStack = [];
    let tagType = TagType.NonStart;
    let attType = AttributeType.None;
    let tagElementName = '';
    let startTagToken = null;
    let preXPathVariable = false;
    let anonymousFunctionParams = false;
    let variableData = null;
    let xsltVariableDeclarations = [];
    let prevToken = null;
    let includeOrImport = false;
    let tagIdentifierName = '';
    let tagAttributeNames = [];
    let tagXmlnsNames = [];
    let rootXmlnsBindings = [];
    let inheritedPrefixes = [];
    let globalVariableData = [];
    let importedGlobalVarNames = new Map();
    let incrementFunctionArity = false;
    let onRootStartTag = true;
    let rootXmlnsName = null;
    let xsltPrefixesToURIs = new Map();
    let namedTemplates = new Map();
    let globalModes = ['#current', '#default'];
    let tagExcludeResultPrefixes = null;
    let requiredLine = position.line;
    let requiredChar = position.character;
    let isOnRequiredToken = false;
    let awaitingRequiredArity = false;
    let keepProcessing = false;
    globalInstructionData.forEach((instruction) => {
        if (instruction.type === xslLexer_1.GlobalInstructionType.Variable || instruction.type === xslLexer_1.GlobalInstructionType.Parameter) {
            if (!importedGlobalVarNames.get(instruction.name)) {
                importedGlobalVarNames.set(instruction.name, instruction);
            }
        }
    });
    importedInstructionData.forEach((instruction) => {
        if (instruction.type === xslLexer_1.GlobalInstructionType.Variable || instruction.type === xslLexer_1.GlobalInstructionType.Parameter) {
            if (!importedGlobalVarNames.get(instruction.name)) {
                importedGlobalVarNames.set(instruction.name, instruction);
            }
        }
    });
    let index = -1;
    for (let token of allTokens) {
        index++;
        lineNumber = token.line;
        let isOnRequiredLine = lineNumber === requiredLine;
        if ((isOnRequiredToken && !keepProcessing) || resultLocation || lineNumber > requiredLine) {
            break;
        }
        isOnRequiredToken = isOnRequiredLine && requiredChar >= token.startCharacter && requiredChar <= (token.startCharacter + token.length);
        let isXMLToken = token.tokenType >= XsltTokenDefinitions.xsltStartTokenNumber;
        if (isXMLToken) {
            inScopeXPathVariablesList = [];
            xpathVariableCurrentlyBeingDefined = false;
            xpathStack = [];
            preXPathVariable = false;
            let xmlCharType = token.charType;
            let xmlTokenType = (token.tokenType - XsltTokenDefinitions.xsltStartTokenNumber);
            switch (xmlTokenType) {
                case xslLexer_1.XSLTokenLevelState.xslElementName:
                    tagElementName = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                    if (tagType === TagType.Start) {
                        tagType = (XsltTokenDefinitions.xslVariable.indexOf(tagElementName) > -1) ? TagType.XSLTvar : TagType.XSLTstart;
                        let xsltToken = token;
                        xsltToken['tagType'] = tagType;
                        startTagToken = token;
                        if (!includeOrImport && tagType !== TagType.XSLTvar && elementStack.length === 1) {
                            includeOrImport = tagElementName === XsltTokenDefinitions.xslImport || tagElementName === XsltTokenDefinitions.xslInclude;
                        }
                    }
                    break;
                case xslLexer_1.XSLTokenLevelState.elementName:
                    tagElementName = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                    if (tagType === TagType.Start) {
                        tagType = TagType.XMLstart;
                        startTagToken = token;
                    }
                    break;
                case xslLexer_1.XSLTokenLevelState.xmlPunctuation:
                    switch (xmlCharType) {
                        case xslLexer_1.XMLCharState.lSt:
                            tagAttributeNames = [];
                            tagXmlnsNames = [];
                            tagIdentifierName = '';
                            variableData = null;
                            tagElementName = '';
                            tagExcludeResultPrefixes = null;
                            tagType = TagType.Start;
                            break;
                        case xslLexer_1.XMLCharState.rStNoAtt:
                        case xslLexer_1.XMLCharState.rSt:
                        case xslLexer_1.XMLCharState.rSelfCt:
                        case xslLexer_1.XMLCharState.rSelfCtNoAtt:
                            // start-tag ended, we're now within the new element scope:
                            if (isXSLT && onRootStartTag) {
                                rootXmlnsBindings.forEach((prefixNsPair) => {
                                    let pfx = prefixNsPair[0];
                                    let namespaceURI = prefixNsPair[1];
                                    let xsltType = functionData_1.FunctionData.namespaces.get(namespaceURI);
                                    if (xsltType !== undefined) {
                                        xsltPrefixesToURIs.set(pfx, xsltType);
                                    }
                                });
                            }
                            onRootStartTag = false;
                            let orginalPrefixes = inheritedPrefixes.slice();
                            tagXmlnsNames.forEach((attName) => {
                                // only need 'xmlns:pfx' - not default xmlns
                                if (attName.length > 6) {
                                    let prefix = attName.substring(6);
                                    if (inheritedPrefixes.indexOf(prefix) < 0) {
                                        inheritedPrefixes.push(prefix);
                                    }
                                }
                            });
                            let problem = false;
                            if (tagExcludeResultPrefixes) {
                                let missingPrefix;
                                if (!(tagExcludeResultPrefixes.prefixes.length === 1 && tagExcludeResultPrefixes.prefixes[0] === '#all')) {
                                    missingPrefix = tagExcludeResultPrefixes.prefixes.find((pfx) => {
                                        if (pfx !== '#default' && inheritedPrefixes.indexOf(pfx) < 0)
                                            return pfx;
                                    });
                                }
                            }
                            if (xmlCharType === xslLexer_1.XMLCharState.rStNoAtt || xmlCharType === xslLexer_1.XMLCharState.rSt) {
                                // on a start tag
                                let inheritedPrefixesCopy = inheritedPrefixes.slice();
                                // if top-level element add global variables - these include following variables also:
                                let newVariablesList = elementStack.length === 0 ? globalVariableData : inScopeVariablesList;
                                //let newVariablesList = inScopeVariablesList;
                                if (variableData !== null) {
                                    if (elementStack.length > 1) {
                                        xsltVariableDeclarations.push(variableData.token);
                                    }
                                    if (startTagToken) {
                                        // if a top-level element, use global variables instad of inScopeVariablesList;
                                        elementStack.push({
                                            namespacePrefixes: inheritedPrefixesCopy, currentVariable: variableData, variables: newVariablesList,
                                            symbolName: tagElementName, symbolID: tagIdentifierName, identifierToken: startTagToken, childSymbols: []
                                        });
                                    }
                                }
                                else if (startTagToken) {
                                    elementStack.push({ namespacePrefixes: inheritedPrefixesCopy, variables: newVariablesList, symbolName: tagElementName, symbolID: tagIdentifierName, identifierToken: startTagToken, childSymbols: [] });
                                }
                                inScopeVariablesList = [];
                                newVariablesList = [];
                                tagType = TagType.NonStart;
                            }
                            else {
                                // self-closed tag: xmlns declarations on this are no longer in scope
                                inheritedPrefixes = orginalPrefixes;
                                if (variableData !== null) {
                                    if (elementStack.length > 1) {
                                        inScopeVariablesList.push(variableData);
                                        xsltVariableDeclarations.push(variableData.token);
                                    }
                                    else {
                                        inScopeVariablesList = globalVariableData;
                                    }
                                }
                                if (startTagToken) {
                                }
                            }
                            break;
                        case xslLexer_1.XMLCharState.rCt:
                            // end of an element close-tag:
                            if (elementStack.length > 0) {
                                let poppedData = elementStack.pop();
                                if (poppedData) {
                                    if (poppedData.symbolName !== tagElementName) {
                                        // not well-nested
                                        if (elementStack.length > 0 && elementStack[elementStack.length - 1].symbolName === tagElementName) {
                                            // recover for benefit of outline view
                                            poppedData = elementStack.pop();
                                        }
                                    }
                                }
                                if (poppedData) {
                                    inheritedPrefixes = poppedData.namespacePrefixes.slice();
                                    inScopeVariablesList = (poppedData) ? poppedData.variables : [];
                                    if (poppedData.currentVariable) {
                                        inScopeVariablesList.push(poppedData.currentVariable);
                                    }
                                }
                            }
                            break;
                    }
                    break;
                case xslLexer_1.XSLTokenLevelState.attributeName:
                case xslLexer_1.XSLTokenLevelState.xmlnsName:
                    rootXmlnsName = null;
                    let attNameText = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                    let problemReported = false;
                    if (!problemReported) {
                        if (tagAttributeNames.indexOf(attNameText) > -1) {
                            //
                        }
                        else {
                            if (xmlTokenType === xslLexer_1.XSLTokenLevelState.xmlnsName) {
                                tagXmlnsNames.push(attNameText);
                                if (onRootStartTag) {
                                    rootXmlnsName = attNameText;
                                }
                            }
                            else {
                                tagAttributeNames.push(attNameText);
                            }
                        }
                    }
                    if (tagType === TagType.XSLTvar) {
                        attType = attNameText === XsltTokenDefinitions.xslNameAtt ? AttributeType.Variable : AttributeType.None;
                    }
                    else if (tagType === TagType.XSLTstart) {
                        if (attNameText === XsltTokenDefinitions.xslNameAtt) {
                            attType = AttributeType.InstructionName;
                        }
                        else if (attNameText === XsltTokenDefinitions.xslModeAtt) {
                            attType = AttributeType.InstructionMode;
                        }
                        else if (attNameText === XsltTokenDefinitions.useAttSet) {
                            attType = AttributeType.UseAttributeSets;
                        }
                        else if (attNameText === XsltTokenDefinitions.excludePrefixes || attNameText === XsltTokenDefinitions.xslExcludePrefixes) {
                            attType = AttributeType.ExcludeResultPrefixes;
                        }
                        else {
                            attType = AttributeType.None;
                        }
                    }
                    else if (attNameText === XsltTokenDefinitions.xslUseAttSet) {
                        attType = AttributeType.UseAttributeSets;
                    }
                    break;
                case xslLexer_1.XSLTokenLevelState.attributeValue:
                    let fullVariableName = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                    let variableName = fullVariableName.substring(1, fullVariableName.length - 1);
                    if (rootXmlnsName !== null) {
                        let prefix = rootXmlnsName.length === 5 ? '' : rootXmlnsName.substr(6);
                        rootXmlnsBindings.push([prefix, variableName]);
                    }
                    switch (attType) {
                        case AttributeType.Variable:
                            tagIdentifierName = variableName;
                            variableData = { token: token, name: variableName };
                            break;
                        case AttributeType.InstructionName:
                            let slashPos = variableName.lastIndexOf('/');
                            if (slashPos > 0) {
                                // package name may be URI
                                variableName = variableName.substring(slashPos + 1);
                            }
                            tagIdentifierName = variableName;
                            if (isOnRequiredToken && tagElementName === 'xsl:call-template') {
                                let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, variableName, xslLexer_1.GlobalInstructionType.Template);
                                resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                            }
                            break;
                        case AttributeType.InstructionMode:
                            if (tagIdentifierName === '') {
                                tagIdentifierName = variableName;
                            }
                            break;
                        case AttributeType.UseAttributeSets:
                            if (isOnRequiredToken) {
                                let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, variableName, xslLexer_1.GlobalInstructionType.AttributeSet);
                                resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                            }
                            break;
                        case AttributeType.ExcludeResultPrefixes:
                            let excludePrefixes = variableName.split(/\s+/);
                            tagExcludeResultPrefixes = { token: token, prefixes: excludePrefixes };
                            break;
                    }
                    attType = AttributeType.None;
                    break;
            }
        }
        else {
            let xpathCharType = token.charType;
            let xpathTokenType = token.tokenType;
            switch (xpathTokenType) {
                case xpLexer_1.TokenLevelState.string:
                    if (xpathStack.length > 0) {
                        let xp = xpathStack[xpathStack.length - 1];
                        if (isOnRequiredToken && (xp.functionArity === 0 && (((_a = xp.function) === null || _a === void 0 ? void 0 : _a.value) === 'key' || ((_b = xp.function) === null || _b === void 0 ? void 0 : _b.value.startsWith('accumulator-'))))) {
                            let keyVal = token.value.substring(1, token.value.length - 1);
                            let instrType = xp.function.value === 'key' ? xslLexer_1.GlobalInstructionType.Key : xslLexer_1.GlobalInstructionType.Accumulator;
                            let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, keyVal, instrType);
                            resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                        }
                        preXPathVariable = xp.preXPathVariable;
                    }
                    break;
                case xpLexer_1.TokenLevelState.function:
                    if (isOnRequiredToken) {
                        awaitingRequiredArity = true;
                        keepProcessing = true;
                    }
                    break;
                case xpLexer_1.TokenLevelState.variable:
                    if ((preXPathVariable && !xpathVariableCurrentlyBeingDefined) || anonymousFunctionParams) {
                        let fullVariableName = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                        let currentVariable = { token: token, name: fullVariableName.substring(1) };
                        if (anonymousFunctionParams) {
                            anonymousFunctionParamList.push(currentVariable);
                            xsltVariableDeclarations.push(token);
                        }
                        else {
                            inScopeXPathVariablesList.push(currentVariable);
                            xpathVariableCurrentlyBeingDefined = true;
                            xsltVariableDeclarations.push(token);
                        }
                    }
                    else {
                        // don't include any current pending variable declarations when resolving
                        if (isOnRequiredToken && !(token.value === '$value' && tagElementName === 'xsl:accumulator-rule')) {
                            let resolvedVariable = XsltTokenDefinitions.resolveXPathVariableReference(document, importedGlobalVarNames, token, xpathVariableCurrentlyBeingDefined, inScopeXPathVariablesList, xpathStack, inScopeVariablesList, elementStack);
                            if (resolvedVariable) {
                                let uri = resolvedVariable.uri ? vscode.Uri.parse(url.pathToFileURL(resolvedVariable.uri).toString()) : document.uri;
                                let startPos = new vscode.Position(resolvedVariable.token.line, resolvedVariable.token.startCharacter);
                                let endPos = new vscode.Position(resolvedVariable.token.line, resolvedVariable.token.startCharacter + resolvedVariable.token.length);
                                resultLocation = new vscode.Location(uri, new vscode.Range(startPos, endPos));
                            }
                        }
                    }
                    break;
                case xpLexer_1.TokenLevelState.complexExpression:
                    let valueText = xsltTokenDiagnostics_1.XsltTokenDiagnostics.getTextForToken(lineNumber, token, document);
                    switch (valueText) {
                        case 'every':
                        case 'for':
                        case 'let':
                        case 'some':
                            preXPathVariable = true;
                            xpathVariableCurrentlyBeingDefined = false;
                            xpathStack.push({ awaitingArity: false, token: token, variables: inScopeXPathVariablesList, preXPathVariable: preXPathVariable, xpathVariableCurrentlyBeingDefined: xpathVariableCurrentlyBeingDefined, isRangeVar: true });
                            break;
                        case 'then':
                            xpathStack.push({ awaitingArity: false, token: token, variables: inScopeXPathVariablesList, preXPathVariable: preXPathVariable, xpathVariableCurrentlyBeingDefined: xpathVariableCurrentlyBeingDefined });
                            inScopeXPathVariablesList = [];
                            break;
                        case 'return':
                        case 'satisfies':
                        case 'else':
                            if (xpathStack.length > 0) {
                                let poppedData = xpathStack.pop();
                                if (poppedData) {
                                    inScopeXPathVariablesList = poppedData.variables;
                                    if (valueText === 'else') {
                                        preXPathVariable = poppedData.preXPathVariable;
                                    }
                                    else {
                                        // todo: if after a return AND a ',' prePathVariable = true; see $pos := $c
                                        preXPathVariable = false;
                                    }
                                    xpathVariableCurrentlyBeingDefined = poppedData.xpathVariableCurrentlyBeingDefined;
                                }
                                else {
                                    inScopeXPathVariablesList = [];
                                    preXPathVariable = false;
                                    xpathVariableCurrentlyBeingDefined = false;
                                }
                            }
                            break;
                    }
                    break;
                case xpLexer_1.TokenLevelState.operator:
                    let functionToken = null;
                    switch (xpathCharType) {
                        case xpLexer_1.CharLevelState.lBr:
                            xpathStack.push({ awaitingArity: false, token: token, variables: inScopeXPathVariablesList, preXPathVariable: preXPathVariable, xpathVariableCurrentlyBeingDefined: xpathVariableCurrentlyBeingDefined });
                            if (anonymousFunctionParams) {
                                // handle case: function($a) {$a + 8} pass params to inside '{...}'				
                                inScopeXPathVariablesList = anonymousFunctionParamList;
                                anonymousFunctionParamList = [];
                                anonymousFunctionParams = false;
                            }
                            else {
                                inScopeXPathVariablesList = [];
                            }
                            preXPathVariable = false;
                            xpathVariableCurrentlyBeingDefined = false;
                            break;
                        case xpLexer_1.CharLevelState.lB:
                            // handle case: function($a)						
                            if (!anonymousFunctionParams && (prevToken === null || prevToken === void 0 ? void 0 : prevToken.tokenType) !== xpLexer_1.TokenLevelState.nodeType) {
                                anonymousFunctionParams = (prevToken === null || prevToken === void 0 ? void 0 : prevToken.tokenType) === xpLexer_1.TokenLevelState.anonymousFunction;
                            }
                            if ((prevToken === null || prevToken === void 0 ? void 0 : prevToken.tokenType) === xpLexer_1.TokenLevelState.function) {
                                functionToken = prevToken;
                            }
                        // intentionally no-break;	
                        case xpLexer_1.CharLevelState.lPr:
                            let xpathItem = { awaitingArity: awaitingRequiredArity, token: token, variables: inScopeXPathVariablesList, preXPathVariable: preXPathVariable, xpathVariableCurrentlyBeingDefined: xpathVariableCurrentlyBeingDefined };
                            awaitingRequiredArity = false;
                            if (functionToken) {
                                xpathItem.function = functionToken;
                                if (incrementFunctionArity) {
                                    xpathItem.functionArity = 1;
                                    incrementFunctionArity = false;
                                }
                                else {
                                    xpathItem.functionArity = 0;
                                }
                            }
                            xpathStack.push(xpathItem);
                            preXPathVariable = false;
                            inScopeXPathVariablesList = [];
                            xpathVariableCurrentlyBeingDefined = false;
                            break;
                        case xpLexer_1.CharLevelState.rB:
                        case xpLexer_1.CharLevelState.rPr:
                        case xpLexer_1.CharLevelState.rBr:
                            if (xpathStack.length > 0) {
                                let poppedData = xpathStack.pop();
                                if (poppedData) {
                                    inScopeXPathVariablesList = poppedData.variables;
                                    preXPathVariable = poppedData.preXPathVariable;
                                    xpathVariableCurrentlyBeingDefined = poppedData.xpathVariableCurrentlyBeingDefined;
                                    if (poppedData.function && poppedData.functionArity !== undefined) {
                                        if ((prevToken === null || prevToken === void 0 ? void 0 : prevToken.charType) !== xpLexer_1.CharLevelState.lB) {
                                            if (poppedData.functionArity !== undefined) {
                                                poppedData.functionArity++;
                                            }
                                        }
                                        if (poppedData.awaitingArity && prevToken) {
                                            keepProcessing = false;
                                            const fnArity = poppedData.functionArity;
                                            const fnName = poppedData.function.value;
                                            let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, fnName, xslLexer_1.GlobalInstructionType.Function, fnArity);
                                            resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                                        }
                                    }
                                }
                                else {
                                    inScopeXPathVariablesList = [];
                                    preXPathVariable = false;
                                    xpathVariableCurrentlyBeingDefined = false;
                                }
                            }
                            break;
                        case xpLexer_1.CharLevelState.sep:
                            if (token.value === ',') {
                                if (xpathStack.length > 0) {
                                    let xp = xpathStack[xpathStack.length - 1];
                                    if (xp.functionArity !== undefined) {
                                        xp.functionArity++;
                                    }
                                    if (xp.isRangeVar) {
                                        preXPathVariable = xp.preXPathVariable;
                                    }
                                }
                                xpathVariableCurrentlyBeingDefined = false;
                            }
                            break;
                        case xpLexer_1.CharLevelState.dSep:
                            if (token.value === '()' && (prevToken === null || prevToken === void 0 ? void 0 : prevToken.tokenType) === xpLexer_1.TokenLevelState.function) {
                                if (awaitingRequiredArity) {
                                    const fnArity = incrementFunctionArity ? 1 : 0;
                                    const fnName = prevToken.value;
                                    let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, fnName, xslLexer_1.GlobalInstructionType.Function, fnArity);
                                    resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                                }
                                awaitingRequiredArity = false;
                                incrementFunctionArity = false;
                            }
                            else if (token.value === '=>') {
                                incrementFunctionArity = true;
                            }
                            break;
                    }
                    break;
                case xpLexer_1.TokenLevelState.functionNameTest:
                    if (isOnRequiredToken) {
                        let { name, arity } = XsltTokenDefinitions.resolveFunctionName(inheritedPrefixes, xsltPrefixesToURIs, token);
                        let instruction = XsltTokenDefinitions.findMatchingDefintion(globalInstructionData, importedInstructionData, name, xslLexer_1.GlobalInstructionType.Function, arity);
                        resultLocation = XsltTokenDefinitions.createLocationFromInstrcution(instruction, document);
                    }
                    break;
            }
        }
        prevToken = token;
    }
    return resultLocation;
};
var XMLPIState;
(function (XMLPIState) {
    XMLPIState[XMLPIState["none"] = 0] = "none";
    XMLPIState[XMLPIState["invalid"] = 1] = "invalid";
    XMLPIState[XMLPIState["Name"] = 2] = "Name";
    XMLPIState[XMLPIState["Eq"] = 3] = "Eq";
    XMLPIState[XMLPIState["Start"] = 4] = "Start";
    XMLPIState[XMLPIState["End"] = 5] = "End";
})(XMLPIState || (XMLPIState = {}));
var XMLPIName;
(function (XMLPIName) {
    XMLPIName[XMLPIName["none"] = 0] = "none";
    XMLPIName[XMLPIName["version"] = 1] = "version";
    XMLPIName[XMLPIName["encoding"] = 2] = "encoding";
    XMLPIName[XMLPIName["standalone"] = 3] = "standalone";
})(XMLPIName || (XMLPIName = {}));
//# sourceMappingURL=xsltTokenDefintions.js.map